/* csvfvc.c                                                vi:se ts=8 sw=8: */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>
#ifdef HAVE_STAT
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#endif

#include "pcrc.h"
#include "crctab.h"
#include "csvrec.h"
#include "fcsvr.h"
#include "fn.h"

static const char *status[] = {
	"Correct", "Bad Rec", "Ill FN", "Missing", 
	"Trunc", "Oversiz", "Bad CRC"
};

int check(struct fcsvr *r)
{
	FILE *stream;
	size_t rd;
	off_t len = 0;
	unsigned long crc = 0xFFFFFFFF;
	int err = 0;
	unsigned char buf[BUFSIZ];

	if (!(stream = fopen(r->fn, strcasecmp(getext(r->fn), ".csv")
	                     ? "rb" : "r")))
		return -3;

	do
		if ((len += rd = fread(buf, 1, sizeof buf, stream)))
			crc = pcrcsum(crc, buf, rd, &pcrctab);
	while (rd == sizeof buf);

	crc = ~crc & 0xFFFFFFFF;

	if (len < r->size)
		err = -4;
	else if (len > r->size)
		err = -5;
	else if (crc != r->crc)
		err = -6;
	
	fclose(stream);

	return err;
}

int check2(struct fcsvr *r, int quick)
{
#ifdef HAVE_STAT
	struct stat st;
#else
	FILE *stream;
#endif
	long len = 0;
	int err = 0;

#ifdef HAVE_STAT
	if (stat(r->fn, &st) == -1)
		return -3;
#else
	if (!(stream = fopen(r->fn, strcasecmp(getext(r->fn), ".csv")
	                     ? "rb" : "r")))
		return -3;
#endif

	if (quick < 2) {
#ifdef HAVE_STAT
		len = st.st_size;
#else
		if (fseek(stream, 0L, SEEK_END) == 0 
		    && (len = ftell(stream)) == -1)
			len = 0;
#endif
	
		if ((off_t)len < r->size)
			err = -4;
		else if ((off_t)len > r->size)
			err = -5;
	}
	
#ifndef HAVE_STAT
	fclose(stream);
#endif

	return err;
}

static char *itoa(char (*buf)[20], int d)
{
	sprintf(*buf, "%d", d);
	return *buf;
}

static int split(const char *s, const char *a, const char *args[], int argl[])
{
	int c = 0;

	while (s && c++ < 15) {
		const char *top = *args++ = s;

		*argl++ = (s = strpbrk(s, a)) ? s++ - top : strlen(top);
	}
	*argl = 0;
	*args = "";

	return c;
}

static void putstr(char **d, char *e, const char *s, int f, int w, int p)
{
	register const char c = f & 010 ? '0' : '_'; /* padding char */
	if (~f & 020) {
		/* Left padding */
		register const char *t = s;
		register int x = p;
		/* Find characters to fill */
		while (*t && (~f & 040 || x-- > 0) && w-- > 0)
			t++;
		/* Output the padding characters */
		while (w-- > 0 && *d < e)
			*(*d)++ = c;
	}
	/* Output string */
	for (; *s && (~f & 040 || p-- > 0); w--)
		if (*d < e)
			switch (f & 03) {
			case 00:
			case 03:
				*(*d)++ = *s++;
				break;
			case 01:
				*(*d)++ = tolower(*s++);
				break;
			case 02:
				*(*d)++ = toupper(*s++);
				break;
			}
		else
			s++;

	if (f & 020) /* Right padding */
		while (w-- > 0 && *d < e)
			*(*d)++ = c;
}

static inline int geti(const char **s)
{
	int r = 0, x;

	if ((x = **s == '-'))
		(*s)++;
	
	while (**s >= '0' && **s <= '9')
		r = *(*s)++ - '0' + r * 10;

	return x ? -r : r;
}

static inline int getu(const char **s)
{
	int r = 0;

	while (**s >= '0' && **s <= '9')
		r = *(*s)++ - '0' + r * 10;

	return r;
}

static inline void getstr(const char **s, char *t, size_t l, const char *acc)
{
	while (**s && strchr(acc, **s))
		if (--l > 0)
			*t++ = *(*s)++;
		else
			(*s)++;
	*t = '\0';
}

static inline void getstr2(const char **s, char *t, size_t l, const char *acc)
{
	while (**s && !strchr(acc, **s))
		if (--l > 0)
			*t++ = *(*s)++;
		else
			(*s)++;
	*t = '\0';
}

static inline int getsub(int x, int l)
{
	/* if (x < 0)
		if (-x < l)
			return l + x;
		else
			return 0;
	else
		if (x < l)
			return x;
		else
			return l; */
	return ((x < 0 ? -x < l : x >= l) ? l : 0) 
		+ ((x < 0 ? -x : x) < l ? x : 0);
}

void refmt(char *d, size_t l, const char *fmt, int recno, const char *fn, 
	const char *ext, const char *desc, const char *path)
{
	char *e = d + l - 1, buf[20], *o;
	int c, t = 0; /* Character, test */

	while ((c = *fmt++) != 0) {
		int f = 0, w = 0, p = 0; /* flags, width, precision */
		int iof = 0, oof = 0, idx = 0, dif = 0, argl[16];
		char accept[16] = "", cond[16] = "";
		const char *args[16];
		
		if (c != '%') {
			if (d < e)
				*d++ = c;
			continue;
		}
		while (~f & 010000)
			switch (*fmt) {
			case '0':
				f |= 010;
				fmt++;
				break;
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				w = getu(&fmt);
				break;
			case '-':
				f |= 020;
				fmt++;
				break;
			case '.':
				f |= 040;
				fmt++;
				p = getu(&fmt);
				break;
			case 'l':
				f |= 01;
				fmt++;
				break;
			case 'u':
				f |= 02;
				fmt++;
				break;
			case 'i': /* ISO 9660 filter */
				f |= 01000;
				fmt++;
				break;
			case 'n': /* numeric conversion */
				f |= 04;
				fmt++;
				break;
			case '#': /* difference */
				fmt++;
				dif = geti(&fmt);
				break;
			case '$': /* word index */
				fmt++;
				idx = geti(&fmt);
				break;
			case '<': /* inner offset */
				fmt++;
				iof = geti(&fmt);
				break;
			case '>': /* outer offset */
				fmt++;
				oof = geti(&fmt);
				break;
			case '/': /* accept list */
				fmt++;
				getstr2(&fmt, accept, sizeof accept, "/");
				fmt += *fmt == '/';
				break;
			case '=': /* Conditional */
				fmt++;
				getstr2(&fmt, cond, sizeof cond, "/");
				fmt += *fmt == '/';
				f |= 0100;
				break;
			case '?':
				fmt++;
				f |= 0200;
				break;
			case ':':
				fmt++;
				f |= 0400;
				break;
			default:
				f |= 010000;
				break;
			}

		o = d;
		switch ((c = *fmt++)) {
			const char *s;
		case '\0':
			fmt--;
			continue;
		case 'd':
			putstr(&d, e, itoa(&buf, recno + dif), f, w, p);
			break;
		case 's':
		case 'f':
		case 'x':
		case 'p':
			s = "";
			switch (c) {
			case 's': s = ext; break;
			case 'f': s = fn; break;
			case 'x': s = desc; break;
			case 'p': s = path; break;
			}
			idx = getsub(idx, split(s + getsub(iof, 
				strlen(s)), accept, args, argl));
			if (f & 04) {
				putstr(&d, e, itoa(&buf, atoi(args[idx] 
					+ oof) + dif), f, w, p);
				break;
			}
			if ((argl[idx] -= oof = getsub(oof, argl[idx])) < p 
			    || ~f & 040)
				p = argl[idx];
			putstr(&d, e, args[idx] + oof, f | 040, w, p);
			break;
		case '%':
			if (d < e)
				*d++ = '%';
			break;
		default:
			if (d < e)
				*d++ = '%';
			if (d < e)
				*d++ = c;
			break;
		}
		if (f & 01000) {
			char *s = o;
			while (s < d)
				if (!isalnum(*s) && *s != '.' && *s != '_')
					*s++ = '_';
				else
					s++;
		}
		if (f & 0100) {
			*d = '\0';
			t = !strcmp(o, cond);
		}
		if ((f & 0200 && !t) || (f & 0400 && t))
			d = o;
	}
	*d++ = '\0';
}


#define FORMAT "%-13s %-13s %-8s %-10s %s\n"

int main(int argc, char *argv[])
{
	struct fcsvr r;
	int i, verb = 0, shf = 0, mode = 0, quick = 0, alt = 0, mach = 0;

	for (i = 1; i < argc; i++) {
		FILE *stream = stdin;
		int err = 0, j, hdr = 0;
		char fmt[64] = "", ext[5] = "", fn[PATH_MAX] = {}, 
			*p = strrchr(argv[i], '/');
		struct {
			int expec, empty;
			unsigned int tfiles, count[7];
			off_t tsize, size[7];
		} st = {getentc(argv[i]), 0, 0, {}, 0, {}};

		p = p && p - argv[i] <= sizeof fn - 5 ? p + 1 : argv[i];

		if (*argv[i] == '-')
			switch (argv[i][1]) {
			case 'v':
				verb++;
				continue;
			case 's':
				shf = 1;
				continue;
			case 'f':
				mode = 1;
				continue;
			case 'd':
				mach = 1;
				continue;
			case 'q':
				quick++;
				continue;
			case 'a':
				alt = 1;
				continue;
			case '\0':
				break;
			default:
				continue;
			}
		else {
			int expec = -1;

			read_hdr(strncpy(fn, argv[i], sizeof fn - 5), 
				&expec, fmt, sizeof fmt, ext, sizeof ext, &r);

			if (st.expec == -1)
				st.expec = expec;
			
			if (!(stream = fopen(argv[i], "r"))) {
				fprintf(stderr, "%s: %s: %s\n", argv[0],
					argv[i], strerror(errno));
				continue;
			}
		}
		
		strncpy(fn, argv[i], p - argv[i]);

		if (!mach) {
			if (!alt)
				printf("CSV file: %s\n", argv[i]);
			else if (mode == 0 && alt < 2) {
				printf("CSV file           Files Empty "
					"Total Verif Bdrec Bdfn  Miss  "
					"Trunc Ovrsz Bdcrc\n"
					"--------           ----- ----- "
					"----- ----- ----- ----  ----  "
					"----- ----- -----\n");
				alt++;
			}
		}


		for (j = 0; (err = getrec(stream, &r)) != -2; j++) {
			char buf[18], *ext2;
			switch (err) {
			case 0:
				if (!r.fn || strchr(r.fn, '/')) {
					err = -2;
					break;
				}
				fn[sizeof fn - 1] = '\0';

				if (*fmt)
					refmt(p - argv[i] + fn, 
						sizeof fn - (p - argv[i]), 
						fmt, j-st.empty+1, r.fn, 
						!*(ext2 = getext(r.fn))
						? ext : ext2, r.desc,
						r.path ? r.path : "");
				else
					strncpy(p - argv[i] + fn, r.fn, 
						sizeof fn - (p - argv[i]) - 1);

				switch (mode) {
					char *datfn, fn2[PATH_MAX];
				case 0:
					datfn = r.fn;
					r.fn = fn;
					err = quick ? check2(&r, quick) 
						: check(&r);
					if (shf)
						r.fn += p - argv[i];
					else
						r.fn = datfn;
					break;
				case 1:
					if (!*fmt)
						break;
					strcpy(fn2, fn);
					strncpy(p - argv[i] + strcpy(fn2, fn), 
						r.fn, sizeof fn2 
						- (p - argv[i]) - 1);
					if (alt)
						printf("%s %s\n", 
							shf ? fn2 : fn,
							shf ? fn : fn2);
					else
						rename(shf ? fn2 : fn,
							shf ? fn : fn2);
					break;
				}
				break;
			case -1:
				sprintf(r.fn = buf, "#%d", j);
				r.size = r.crc = 0;
				r.desc = NULL;
				break;
			case -3:
				st.empty++;
				continue;
			}
			if (mode)
				continue;

			st.tfiles++;
			st.tsize += r.size;
			st.count[-err]++;
			st.size[-err] += r.size;
			
			if (mach && err < 0)
				printf("%lx %lu %s%s\n", r.crc, r.size, r.path 
					&& !alt ? r.path : "", 
					r.fn ? r.fn : "");
			else if (!mach && ((verb > 0 && err < 0) || verb > 1)) {
				if (!alt && !hdr) {
					printf("\n" FORMAT FORMAT,
						"Status", "File name", "Size",
						"CRC 32", "Description",
						"------", "---------", "----", 
						"------", "-----------");
					hdr = 1;
				}
				printf("%-13s %-13s %-8ld %08lx   %.*s%s\n", 
					status[-err], r.fn ? r.fn : "", 
					r.size, r.crc, strlen(r.desc 
					? r.desc : "") <= 32 ? 32 : 29, 
					r.desc ? r.desc : "", strlen(r.desc 
					? r.desc : "") <= 32 ? "" : "...");
			}
		}
		
		if (mode == 0 && alt && !mach) {
			printf("%-18s %5d %5d %5d %5d %5d "
				"%5d %5d %5d %5d %5d\n",
				argv[i], st.tfiles, st.empty, st.expec, 
				st.count[0], st.count[1], st.count[2], 
				st.count[3], st.count[4], st.count[5],
				st.count[6]);
		}
		if (mode || alt || mach) {
			if (stream != stdin)
				fclose(stream);
			continue;
		}

		printf("\nResults for collection:\n-----------------------\n"
			"Files in collection: %5u", st.tfiles);

		if (st.expec != -1 && st.expec != st.tfiles)
			printf(" of %5d", st.expec);
		if (st.empty)
			printf(", %5d empty", st.empty);

		printf(" (%lu bytes)\n\n", st.tsize);

		for (j = 0; j < 7; j++)
			if (st.count[j])
				printf("%s:%.*s %5d (%3d.%02d%%)   "
					"(%lu bytes)\n", status[j],
					7 - (int)strlen(status[j]), "       ",
					st.count[j], 
					st.tfiles ? st.count[j] * 100 
					/ st.tfiles : 0, st.tfiles 
					? st.count[j] * 10000 / st.tfiles
					% 100 : 0, st.size[j]);

		if (i < argc - 1)
			printf("\n\n");

		if (stream != stdin)
			fclose(stream);
	}
	return EXIT_SUCCESS;
}
